@@ -1,4 +1,4 @@
-use crate::{command::Action, utils::new_table_state};
+use crate::command::Action;
 
 use super::*;
 
@@ -7,86 +7,91 @@ pub fn handle_key_sequence_for_popup(
     key_sequence: &KeySequence,
     send: &mpsc::Sender<ClientRequest>,
     state: &SharedState,
-    ui: &mut UIStateGuard,
 ) -> Result<bool> {
+    let ui = state.ui.lock();
+
     match ui.popup.as_ref().unwrap() {
-        PopupState::ContextSearch(_) => {
-            handle_key_sequence_for_search_popup(key_sequence, send, state, ui)
+        PopupState::Search { .. } => {
+            drop(ui);
+            handle_key_sequence_for_search_popup(key_sequence, send, state)
         }
-        PopupState::ArtistList(..) => handle_key_sequence_for_list_popup(
-            key_sequence,
-            state,
-            ui,
-            match ui.popup {
-                Some(PopupState::ArtistList(ref artists, _)) => artists.len(),
-                _ => unreachable!(),
-            },
-            |_, _| {},
-            |ui: &mut UIStateGuard, id: usize| -> Result<()> {
-                let artists = match ui.popup {
-                    Some(PopupState::ArtistList(ref artists, _)) => artists,
-                    _ => unreachable!(),
-                };
+        PopupState::ArtistList(artists, _) => {
+            let n_items = artists.len();
 
-                let context_id = ContextId::Artist(artists[id].id.clone());
-                send.send(ClientRequest::GetContext(context_id.clone()))?;
-                ui.new_page(PageState::Browsing(context_id));
+            drop(ui);
+            handle_key_sequence_for_list_popup(
+                key_sequence,
+                state,
+                n_items,
+                |_, _| {},
+                |ui: &mut UIStateGuard, id: usize| -> Result<()> {
+                    let artists = match ui.popup {
+                        Some(PopupState::ArtistList(ref artists, _)) => artists,
+                        _ => return Ok(()),
+                    };
 
-                Ok(())
-            },
-            |ui: &mut UIStateGuard| {
-                ui.popup = None;
-            },
-        ),
-        PopupState::UserPlaylistList(action, playlists, _) => {
+                    let context_id = ContextId::Artist(artists[id].id.clone());
+                    ui.create_new_page(PageState::Context(
+                        None,
+                        ContextPageType::Browsing(context_id),
+                    ));
+
+                    Ok(())
+                },
+                |ui: &mut UIStateGuard| {
+                    ui.popup = None;
+                },
+            )
+        }
+        PopupState::UserPlaylistList(action, _) => {
             match action {
                 PlaylistPopupAction::Browse => {
-                    let playlist_uris = playlists.iter().map(|p| p.id.uri()).collect::<Vec<_>>();
+                    let playlist_uris = state
+                        .data
+                        .read()
+                        .user_data
+                        .playlists
+                        .iter()
+                        .map(|p| p.id.uri())
+                        .collect::<Vec<_>>();
 
+                    drop(ui);
                     handle_key_sequence_for_context_browsing_list_popup(
                         key_sequence,
-                        send,
                         state,
-                        ui,
                         playlist_uris,
-                        model::Type::Playlist,
+                        rspotify_model::Type::Playlist,
                     )
                 }
-                PlaylistPopupAction::AddTrack(ref track_id) => {
+                PlaylistPopupAction::AddTrack(track_id) => {
                     let track_id = track_id.clone();
+                    let playlist_ids = state
+                        .data
+                        .read()
+                        .user_data
+                        .playlists_created_by_user()
+                        .into_iter()
+                        .map(|p| p.id.clone())
+                        .collect::<Vec<_>>();
 
+                    drop(ui);
                     handle_key_sequence_for_list_popup(
                         key_sequence,
                         state,
-                        ui,
-                        {
-                            match ui.popup {
-                                Some(PopupState::UserPlaylistList(_, ref playlists, _)) => {
-                                    playlists.len()
-                                }
-                                _ => unreachable!(),
-                            }
-                        },
+                        playlist_ids.len(),
                         |_, _| {},
                         |ui: &mut UIStateGuard, id: usize| -> Result<()> {
-                            let playlists = match ui.popup {
-                                Some(PopupState::UserPlaylistList(_, ref playlists, _)) => {
-                                    playlists
-                                }
-                                _ => unreachable!(),
-                            };
-
                             // when adding a new track to a playlist, we need to remove
                             // the cache for that playlist
                             state
-                                .player
+                                .data
                                 .write()
-                                .unwrap()
-                                .context_cache
-                                .pop(&playlists[id].id.uri());
+                                .caches
+                                .context
+                                .pop(&playlist_ids[id].uri());
 
                             send.send(ClientRequest::AddTrackToPlaylist(
-                                playlists[id].id.clone(),
+                                playlist_ids[id].clone(),
                                 track_id.clone(),
                             ))?;
                             ui.popup = None;
@@ -100,72 +105,75 @@ pub fn handle_key_sequence_for_popup(
             }
         }
         PopupState::UserFollowedArtistList(_) => {
-            let player = state.player.read().unwrap();
-            let artist_uris = player
-                .user_followed_artists
+            let artist_uris = state
+                .data
+                .read()
+                .user_data
+                .followed_artists
                 .iter()
                 .map(|a| a.id.uri())
                 .collect::<Vec<_>>();
 
+            drop(ui);
             handle_key_sequence_for_context_browsing_list_popup(
                 key_sequence,
-                send,
                 state,
-                ui,
                 artist_uris,
-                model::Type::Artist,
+                rspotify_model::Type::Artist,
             )
         }
         PopupState::UserSavedAlbumList(_) => {
-            let player = state.player.read().unwrap();
-            let album_uris = player
-                .user_saved_albums
+            let album_uris = state
+                .data
+                .read()
+                .user_data
+                .saved_albums
                 .iter()
                 .map(|a| a.id.uri())
                 .collect::<Vec<_>>();
 
+            drop(ui);
             handle_key_sequence_for_context_browsing_list_popup(
                 key_sequence,
-                send,
                 state,
-                ui,
                 album_uris,
-                model::Type::Album,
+                rspotify_model::Type::Album,
             )
         }
-        PopupState::ThemeList(_, _) => handle_key_sequence_for_list_popup(
-            key_sequence,
-            state,
-            ui,
-            match ui.popup {
-                Some(PopupState::ThemeList(ref themes, _)) => themes.len(),
-                _ => unreachable!(),
-            },
-            |ui: &mut UIStateGuard, id: usize| {
-                ui.theme = match ui.popup {
-                    Some(PopupState::ThemeList(ref themes, _)) => themes[id].clone(),
-                    _ => unreachable!(),
-                };
-            },
-            |ui: &mut UIStateGuard, _| -> Result<()> {
-                ui.popup = None;
-                Ok(())
-            },
-            |ui: &mut UIStateGuard| {
-                ui.theme = match ui.popup {
-                    Some(PopupState::ThemeList(ref themes, _)) => themes[0].clone(),
-                    _ => unreachable!(),
-                };
-                ui.popup = None;
-            },
-        ),
-        PopupState::DeviceList(_) => {
-            let player = state.player.read().unwrap();
+        PopupState::ThemeList(themes, _) => {
+            let n_items = themes.len();
 
+            drop(ui);
+            handle_key_sequence_for_list_popup(
+                key_sequence,
+                state,
+                n_items,
+                |ui: &mut UIStateGuard, id: usize| {
+                    ui.theme = match ui.popup {
+                        Some(PopupState::ThemeList(ref themes, _)) => themes[id].clone(),
+                        _ => return,
+                    };
+                },
+                |ui: &mut UIStateGuard, _| -> Result<()> {
+                    ui.popup = None;
+                    Ok(())
+                },
+                |ui: &mut UIStateGuard| {
+                    ui.theme = match ui.popup {
+                        Some(PopupState::ThemeList(ref themes, _)) => themes[0].clone(),
+                        _ => return,
+                    };
+                    ui.popup = None;
+                },
+            )
+        }
+        PopupState::DeviceList(_) => {
+            let player = state.player.read();
+
+            drop(ui);
             handle_key_sequence_for_list_popup(
                 key_sequence,
                 state,
-                ui,
                 player.devices.len(),
                 |_, _| {},
                 |ui: &mut UIStateGuard, id: usize| -> Result<()> {
@@ -181,11 +189,14 @@ pub fn handle_key_sequence_for_popup(
                 },
             )
         }
-        PopupState::CommandHelp(_) => {
-            handle_key_sequence_for_command_help_popup(key_sequence, state, ui)
+        PopupState::CommandHelp { .. } => {
+            drop(ui);
+            handle_key_sequence_for_command_help_popup(key_sequence, state)
         }
-        PopupState::ActionList(ref item, _) => {
-            handle_key_sequence_for_action_list_popup(item.actions(), key_sequence, send, state, ui)
+        PopupState::ActionList(item, ..) => {
+            let actions = item.actions();
+            drop(ui);
+            handle_key_sequence_for_action_list_popup(actions, key_sequence, send, state)
         }
     }
 }
@@ -195,11 +206,13 @@ fn handle_key_sequence_for_search_popup(
     key_sequence: &KeySequence,
     send: &mpsc::Sender<ClientRequest>,
     state: &SharedState,
-    ui: &mut UIStateGuard,
 ) -> Result<bool> {
+    let mut ui = state.ui.lock();
+
+    // handle user's input that updates the search query
     let query = match ui.popup {
-        Some(PopupState::ContextSearch(ref mut query)) => query,
-        _ => unreachable!(),
+        Some(PopupState::Search { ref mut query }) => query,
+        _ => return Ok(false),
     };
     if key_sequence.keys.len() == 1 {
         if let Key::None(c) = key_sequence.keys[0] {
@@ -230,30 +243,26 @@ fn handle_key_sequence_for_search_popup(
             Command::ClosePopup => {
                 ui.window.select(Some(0));
                 ui.popup = None;
+                Ok(true)
             }
             _ => match ui.current_page() {
+                PageState::Library => {
+                    drop(ui);
+                    window::handle_key_sequence_for_library_window(key_sequence, state)
+                }
                 PageState::Recommendations(..) => {
-                    return window::handle_key_sequence_for_recommendation_window(
-                        key_sequence,
-                        send,
-                        state,
-                        ui,
-                    );
+                    drop(ui);
+                    window::handle_key_sequence_for_recommendation_window(key_sequence, send, state)
                 }
-                PageState::Browsing(_) | PageState::CurrentPlaying => {
-                    return window::handle_key_sequence_for_context_window(
-                        key_sequence,
-                        send,
-                        state,
-                        ui,
-                    );
+                PageState::Context(..) => {
+                    drop(ui);
+                    window::handle_key_sequence_for_context_window(key_sequence, send, state)
                 }
-                _ => {}
+                PageState::Searching { .. } => Ok(false),
             },
         },
-        None => return Ok(false),
+        None => Ok(false),
     }
-    Ok(true)
 }
 
 /// handles a key sequence for a context list popup in which
@@ -265,32 +274,30 @@ fn handle_key_sequence_for_search_popup(
 /// - `uri_type`: an enum represents the type of a context in the list (`playlist`, `artist`, etc)
 fn handle_key_sequence_for_context_browsing_list_popup(
     key_sequence: &KeySequence,
-    send: &mpsc::Sender<ClientRequest>,
     state: &SharedState,
-    ui: &mut UIStateGuard,
     uris: Vec<String>,
-    context_type: model::Type,
+    context_type: rspotify_model::Type,
 ) -> Result<bool> {
     handle_key_sequence_for_list_popup(
         key_sequence,
         state,
-        ui,
         uris.len(),
         |_, _| {},
         |ui: &mut UIStateGuard, id: usize| -> Result<()> {
             let uri = uris[id].clone();
             let context_id = match context_type {
-                model::Type::Playlist => ContextId::Playlist(PlaylistId::from_uri(&uri)?),
-                model::Type::Artist => ContextId::Artist(ArtistId::from_uri(&uri)?),
-                model::Type::Album => ContextId::Album(AlbumId::from_uri(&uri)?),
+                rspotify_model::Type::Playlist => ContextId::Playlist(PlaylistId::from_uri(&uri)?),
+                rspotify_model::Type::Artist => ContextId::Artist(ArtistId::from_uri(&uri)?),
+                rspotify_model::Type::Album => ContextId::Album(AlbumId::from_uri(&uri)?),
                 _ => {
                     return Ok(());
                 }
             };
 
-            send.send(ClientRequest::GetContext(context_id.clone()))?;
-
-            ui.new_page(PageState::Browsing(context_id));
+            ui.create_new_page(PageState::Context(
+                None,
+                ContextPageType::Browsing(context_id),
+            ));
 
             Ok(())
         },
@@ -311,7 +318,6 @@ fn handle_key_sequence_for_context_browsing_list_popup(
 fn handle_key_sequence_for_list_popup(
     key_sequence: &KeySequence,
     state: &SharedState,
-    ui: &mut UIStateGuard,
     n_items: usize,
     on_select_func: impl Fn(&mut UIStateGuard, usize),
     on_choose_func: impl Fn(&mut UIStateGuard, usize) -> Result<()>,
@@ -325,6 +331,8 @@ fn handle_key_sequence_for_list_popup(
         None => return Ok(false),
     };
 
+    let mut ui = state.ui.lock();
+
     let popup = ui.popup.as_mut().unwrap();
     let current_id = popup.list_selected().unwrap();
 
@@ -332,20 +340,20 @@ fn handle_key_sequence_for_list_popup(
         Command::SelectPreviousOrScrollUp => {
             if current_id > 0 {
                 popup.list_select(Some(current_id - 1));
-                on_select_func(ui, current_id - 1);
+                on_select_func(&mut ui, current_id - 1);
             }
         }
         Command::SelectNextOrScrollDown => {
             if current_id + 1 < n_items {
                 popup.list_select(Some(current_id + 1));
-                on_select_func(ui, current_id + 1);
+                on_select_func(&mut ui, current_id + 1);
             }
         }
         Command::ChooseSelected => {
-            on_choose_func(ui, current_id)?;
+            on_choose_func(&mut ui, current_id)?;
         }
         Command::ClosePopup => {
-            on_close_func(ui);
+            on_close_func(&mut ui);
         }
         _ => {
             return Ok(false);
@@ -358,7 +366,6 @@ fn handle_key_sequence_for_list_popup(
 fn handle_key_sequence_for_command_help_popup(
     key_sequence: &KeySequence,
     state: &SharedState,
-    ui: &mut UIStateGuard,
 ) -> Result<bool> {
     let command = match state
         .keymap_config
@@ -368,9 +375,11 @@ fn handle_key_sequence_for_command_help_popup(
         None => return Ok(false),
     };
 
+    let mut ui = state.ui.lock();
+
     let offset = match ui.popup {
-        Some(PopupState::CommandHelp(ref mut offset)) => offset,
-        _ => unreachable!(),
+        Some(PopupState::CommandHelp { ref mut offset }) => offset,
+        _ => return Ok(false),
     };
     match command {
         Command::ClosePopup => {
@@ -395,18 +404,16 @@ fn handle_key_sequence_for_action_list_popup(
     key_sequence: &KeySequence,
     send: &mpsc::Sender<ClientRequest>,
     state: &SharedState,
-    ui: &mut UIStateGuard,
 ) -> Result<bool> {
     handle_key_sequence_for_list_popup(
         key_sequence,
         state,
-        ui,
         actions.len(),
         |_, _| {},
         |ui: &mut UIStateGuard, id: usize| -> Result<()> {
             let item = match ui.popup {
-                Some(PopupState::ActionList(ref item, _)) => item,
-                _ => unreachable!(),
+                Some(PopupState::ActionList(ref item, ..)) => item,
+                _ => return Ok(()),
             };
 
             match item {
@@ -415,8 +422,10 @@ fn handle_key_sequence_for_action_list_popup(
                         if let Some(ref album) = track.album {
                             let uri = album.id.uri();
                             let context_id = ContextId::Album(AlbumId::from_uri(&uri)?);
-                            send.send(ClientRequest::GetContext(context_id.clone()))?;
-                            ui.new_page(PageState::Browsing(context_id));
+                            ui.create_new_page(PageState::Context(
+                                None,
+                                ContextPageType::Browsing(context_id),
+                            ));
                         }
                     }
                     Action::BrowseArtist => {
@@ -426,21 +435,11 @@ fn handle_key_sequence_for_action_list_popup(
                         ));
                     }
                     Action::AddTrackToPlaylist => {
-                        let player = state.player.read().unwrap();
-                        if let Some(ref user) = player.user {
-                            let playlists = player
-                                .user_playlists
-                                .iter()
-                                .filter(|p| p.owner.1 == user.id)
-                                .cloned()
-                                .collect();
-
-                            ui.popup = Some(PopupState::UserPlaylistList(
-                                PlaylistPopupAction::AddTrack(track.id.clone()),
-                                playlists,
-                                new_list_state(),
-                            ));
-                        }
+                        send.send(ClientRequest::GetUserPlaylists)?;
+                        ui.popup = Some(PopupState::UserPlaylistList(
+                            PlaylistPopupAction::AddTrack(track.id.clone()),
+                            new_list_state(),
+                        ));
                     }
                     Action::SaveToLibrary => {
                         send.send(ClientRequest::SaveToLibrary(item.clone()))?;
@@ -449,8 +448,7 @@ fn handle_key_sequence_for_action_list_popup(
                     Action::BrowseRecommendations => {
                         let seed = SeedItem::Track(track.clone());
                         send.send(ClientRequest::GetRecommendations(seed.clone()))?;
-                        ui.new_page(PageState::Recommendations(seed, None));
-                        ui.window = WindowState::Recommendations(new_table_state());
+                        ui.create_new_page(PageState::Recommendations(seed));
                     }
                 },
                 Item::Album(album) => match actions[id] {
@@ -474,8 +472,7 @@ fn handle_key_sequence_for_action_list_popup(
                     Action::BrowseRecommendations => {
                         let seed = SeedItem::Artist(artist.clone());
                         send.send(ClientRequest::GetRecommendations(seed.clone()))?;
-                        ui.new_page(PageState::Recommendations(seed, None));
-                        ui.window = WindowState::Recommendations(new_table_state());
+                        ui.create_new_page(PageState::Recommendations(seed));
                     }
                     _ => {}
                 },
