@@ -5,7 +5,6 @@ pub fn handle_key_sequence_for_context_window(
     key_sequence: &KeySequence,
     send: &mpsc::Sender<ClientRequest>,
     state: &SharedState,
-    ui: &mut UIStateGuard,
 ) -> Result<bool> {
     let command = match state
         .keymap_config
@@ -15,6 +14,7 @@ pub fn handle_key_sequence_for_context_window(
         None => return Ok(false),
     };
 
+    let mut ui = state.ui.lock();
     match command {
         Command::FocusNextWindow => {
             ui.window.next();
@@ -22,81 +22,83 @@ pub fn handle_key_sequence_for_context_window(
         Command::FocusPreviousWindow => {
             ui.window.previous();
         }
-        Command::SearchContext => {
+        Command::Search => {
             ui.window.select(Some(0));
-            ui.popup = Some(PopupState::ContextSearch("".to_owned()));
+            ui.popup = Some(PopupState::Search {
+                query: "".to_owned(),
+            });
         }
         Command::PlayRandom => {
-            let player = state.player.read().unwrap();
-            let context = player.context();
+            if let Some(uri) = ui.current_page().context_uri() {
+                let data = state.data.read();
 
-            // randomly play a track from the current context
-            if let Some(context) = context {
-                let tracks = context.tracks();
-                let offset = match context {
-                    // Spotify does not allow to manually specify `offset` for artist context
-                    Context::Artist(..) => None,
-                    _ => {
-                        let id = rand::thread_rng().gen_range(0..tracks.len());
-                        Some(model::Offset::for_uri(&tracks[id].id.uri()))
-                    }
-                };
+                // randomly play a track from the current context
+                if let Some(context) = data.caches.context.peek(&uri) {
+                    let tracks = context.tracks();
+                    let offset = match context {
+                        // Spotify does not allow to manually specify `offset` for artist context
+                        Context::Artist { .. } => None,
+                        _ => {
+                            let id = rand::thread_rng().gen_range(0..tracks.len());
+                            Some(rspotify_model::Offset::for_uri(&tracks[id].id.uri()))
+                        }
+                    };
 
-                send.send(ClientRequest::Player(PlayerRequest::StartPlayback(
-                    Playback::Context(player.context_id.clone().unwrap(), offset),
-                )))?;
+                    let context_id = match ui.current_page() {
+                        PageState::Context(context_id, _) => context_id.clone().unwrap(),
+                        _ => return Ok(false),
+                    };
+
+                    send.send(ClientRequest::Player(PlayerRequest::StartPlayback(
+                        Playback::Context(context_id, offset),
+                    )))?;
+                }
             }
         }
         _ => {
-            // handles sort/reverse commands separately
-            if state.player.read().unwrap().context().is_some() {
-                let sort_order = match command {
-                    Command::SortTrackByTitle => Some(ContextSortOrder::TrackName),
-                    Command::SortTrackByAlbum => Some(ContextSortOrder::Album),
-                    Command::SortTrackByArtists => Some(ContextSortOrder::Artists),
-                    Command::SortTrackByAddedDate => Some(ContextSortOrder::AddedAt),
-                    Command::SortTrackByDuration => Some(ContextSortOrder::Duration),
-                    _ => None,
-                };
-                match sort_order {
-                    Some(sort_order) => {
-                        state
-                            .player
-                            .write()
-                            .unwrap()
-                            .context_mut()
-                            .unwrap()
-                            .sort_tracks(sort_order);
-                        return Ok(true);
-                    }
-                    None => {
-                        if command == Command::ReverseTrackOrder {
-                            state
-                                .player
-                                .write()
-                                .unwrap()
-                                .context_mut()
-                                .unwrap()
-                                .reverse_tracks();
-                            return Ok(true);
-                        }
+            // handle sort/reverse tracks commands
+            let order = match command {
+                Command::SortTrackByTitle => Some(TrackOrder::TrackName),
+                Command::SortTrackByAlbum => Some(TrackOrder::Album),
+                Command::SortTrackByArtists => Some(TrackOrder::Artists),
+                Command::SortTrackByAddedDate => Some(TrackOrder::AddedAt),
+                Command::SortTrackByDuration => Some(TrackOrder::Duration),
+                _ => None,
+            };
+
+            if let Some(order) = order {
+                if let Some(uri) = ui.current_page().context_uri() {
+                    drop(ui);
+                    let mut data = state.data.write();
+                    if let Some(context) = data.caches.context.peek_mut(&uri) {
+                        context.sort_tracks(order);
                     }
                 }
+                return Ok(true);
+            }
+            if command == Command::ReverseTrackOrder {
+                if let Some(uri) = ui.current_page().context_uri() {
+                    drop(ui);
+                    let mut data = state.data.write();
+                    if let Some(context) = data.caches.context.peek_mut(&uri) {
+                        context.reverse_tracks();
+                    }
+                }
+                return Ok(true);
             }
 
             // the command hasn't been handled, assign the job to the focused subwindow's handler
-            return handle_command_for_focused_context_subwindow(command, send, ui, state);
+            drop(ui);
+            return handle_command_for_focused_context_subwindow(command, send, state);
         }
     }
     Ok(true)
 }
 
-/// handles a key sequence for a recommendation window
-pub fn handle_key_sequence_for_recommendation_window(
+/// handles a key sequence for a library window
+pub fn handle_key_sequence_for_library_window(
     key_sequence: &KeySequence,
-    send: &mpsc::Sender<ClientRequest>,
     state: &SharedState,
-    ui: &mut UIStateGuard,
 ) -> Result<bool> {
     let command = match state
         .keymap_config
@@ -106,22 +108,93 @@ pub fn handle_key_sequence_for_recommendation_window(
         None => return Ok(false),
     };
 
-    let tracks = match ui.current_page() {
-        PageState::Recommendations(_, tracks) => tracks.clone().unwrap_or_default(),
-        _ => unreachable!(),
+    let focus_state = match state.ui.lock().window {
+        WindowState::Library { focus, .. } => focus,
+        _ => return Ok(false),
     };
 
     match command {
-        Command::SearchContext => {
+        Command::FocusNextWindow => {
+            state.ui.lock().window.next();
+            Ok(true)
+        }
+        Command::FocusPreviousWindow => {
+            state.ui.lock().window.previous();
+            Ok(true)
+        }
+        Command::Search => {
+            let mut ui = state.ui.lock();
             ui.window.select(Some(0));
-            ui.popup = Some(PopupState::ContextSearch("".to_owned()));
+            ui.popup = Some(PopupState::Search {
+                query: "".to_owned(),
+            });
+            Ok(true)
+        }
+        _ => {
+            let data = state.data.read();
+            match focus_state {
+                LibraryFocusState::Playlists => handle_command_for_playlist_list_subwindow(
+                    command,
+                    state,
+                    state.filtered_items_by_search(&data.user_data.playlists),
+                ),
+                LibraryFocusState::SavedAlbums => handle_command_for_album_list_subwindow(
+                    command,
+                    state,
+                    state.filtered_items_by_search(&data.user_data.saved_albums),
+                ),
+                LibraryFocusState::FollowedArtists => handle_command_for_artist_list_subwindow(
+                    command,
+                    state,
+                    state.filtered_items_by_search(&data.user_data.followed_artists),
+                ),
+            }
+        }
+    }
+}
+
+/// handles a key sequence for a recommendation window
+pub fn handle_key_sequence_for_recommendation_window(
+    key_sequence: &KeySequence,
+    send: &mpsc::Sender<ClientRequest>,
+    state: &SharedState,
+) -> Result<bool> {
+    let command = match state
+        .keymap_config
+        .find_command_from_key_sequence(key_sequence)
+    {
+        Some(command) => command,
+        None => return Ok(false),
+    };
+
+    let data = state.data.read();
+
+    // get the recommendation tracks from the cache
+    let seed_uri = match state.ui.lock().current_page() {
+        PageState::Recommendations(seed) => seed.uri(),
+        _ => return Ok(false),
+    };
+    let tracks = data
+        .caches
+        .recommendation
+        .peek(&seed_uri)
+        .map(|tracks| state.filtered_items_by_search(tracks))
+        .unwrap_or_default();
+
+    match command {
+        Command::Search => {
+            let mut ui = state.ui.lock();
+            ui.window.select(Some(0));
+            ui.popup = Some(PopupState::Search {
+                query: "".to_owned(),
+            });
             Ok(true)
         }
         Command::PlayRandom => {
             // randomly play a song from the list of recommendation tracks
             let offset = {
                 let id = rand::thread_rng().gen_range(0..tracks.len());
-                Some(model::Offset::for_uri(&tracks[id].id.uri()))
+                Some(rspotify_model::Offset::for_uri(&tracks[id].id.uri()))
             };
             send.send(ClientRequest::Player(PlayerRequest::StartPlayback(
                 Playback::URIs(tracks.iter().map(|t| t.id.clone()).collect(), offset),
@@ -132,10 +205,10 @@ pub fn handle_key_sequence_for_recommendation_window(
         _ => handle_command_for_track_table_subwindow(
             command,
             send,
-            ui,
+            state,
             None,
             Some(tracks.iter().map(|t| &t.id).collect()),
-            ui.filtered_items_by_search(&tracks),
+            tracks,
         ),
     }
 }
@@ -145,20 +218,22 @@ pub fn handle_key_sequence_for_search_window(
     key_sequence: &KeySequence,
     send: &mpsc::Sender<ClientRequest>,
     state: &SharedState,
-    ui: &mut UIStateGuard,
 ) -> Result<bool> {
+    let mut ui = state.ui.lock();
+
     let focus_state = match ui.window {
-        WindowState::Search(_, _, _, _, focus) => focus,
+        WindowState::Search { focus, .. } => focus,
         _ => {
             return Ok(false);
         }
     };
 
-    let (query, search_results) = match ui.current_page_mut() {
-        PageState::Searching(ref mut query, ref mut search_results) => {
-            (query, search_results.clone())
-        }
-        _ => unreachable!(),
+    let (input, current_query) = match ui.current_page_mut() {
+        PageState::Searching {
+            input,
+            current_query,
+        } => (input, current_query),
+        _ => return Ok(false),
     };
 
     // handle user's input
@@ -167,18 +242,19 @@ pub fn handle_key_sequence_for_search_window(
             if let Key::None(c) = key_sequence.keys[0] {
                 match c {
                     KeyCode::Char(c) => {
-                        query.push(c);
+                        input.push(c);
                         return Ok(true);
                     }
                     KeyCode::Backspace => {
-                        if !query.is_empty() {
-                            query.pop().unwrap();
+                        if !input.is_empty() {
+                            input.pop().unwrap();
                         }
                         return Ok(true);
                     }
                     KeyCode::Enter => {
-                        if !query.is_empty() {
-                            send.send(ClientRequest::Search(query.clone()))?;
+                        if !input.is_empty() {
+                            *current_query = input.clone();
+                            send.send(ClientRequest::Search(input.clone()))?;
                         }
                         return Ok(true);
                     }
@@ -196,6 +272,9 @@ pub fn handle_key_sequence_for_search_window(
         None => return Ok(false),
     };
 
+    let data = state.data.read();
+    let search_results = data.caches.search.peek(current_query);
+
     match command {
         Command::FocusNextWindow => {
             ui.window.next();
@@ -207,25 +286,37 @@ pub fn handle_key_sequence_for_search_window(
         }
         // determine the current focused subwindow inside the search window,
         // and assign the handling job to the corresponding handler
-        _ => match focus_state {
-            SearchFocusState::Input => Ok(false),
-            SearchFocusState::Tracks => {
-                let tracks = search_results.tracks.iter().collect::<Vec<_>>();
-                handle_command_for_track_list_subwindow(command, send, ui, tracks)
+        _ => {
+            drop(ui);
+
+            match focus_state {
+                SearchFocusState::Input => Ok(false),
+                SearchFocusState::Tracks => {
+                    let tracks = search_results
+                        .map(|s| s.tracks.iter().collect())
+                        .unwrap_or_default();
+                    handle_command_for_track_list_subwindow(command, send, state, tracks)
+                }
+                SearchFocusState::Artists => {
+                    let artists = search_results
+                        .map(|s| s.artists.iter().collect())
+                        .unwrap_or_default();
+                    handle_command_for_artist_list_subwindow(command, state, artists)
+                }
+                SearchFocusState::Albums => {
+                    let albums = search_results
+                        .map(|s| s.albums.iter().collect())
+                        .unwrap_or_default();
+                    handle_command_for_album_list_subwindow(command, state, albums)
+                }
+                SearchFocusState::Playlists => {
+                    let playlists = search_results
+                        .map(|s| s.playlists.iter().collect())
+                        .unwrap_or_default();
+                    handle_command_for_playlist_list_subwindow(command, state, playlists)
+                }
             }
-            SearchFocusState::Artists => {
-                let artists = search_results.artists.iter().collect::<Vec<_>>();
-                handle_command_for_artist_list_subwindow(command, send, ui, artists)
-            }
-            SearchFocusState::Albums => {
-                let albums = search_results.albums.iter().collect::<Vec<_>>();
-                handle_command_for_album_list_subwindow(command, send, ui, albums)
-            }
-            SearchFocusState::Playlists => {
-                let playlists = search_results.playlists.iter().collect::<Vec<_>>();
-                handle_command_for_playlist_list_subwindow(command, send, ui, playlists)
-            }
-        },
+        }
     }
 }
 
@@ -236,58 +327,63 @@ pub fn handle_key_sequence_for_search_window(
 pub fn handle_command_for_focused_context_subwindow(
     command: Command,
     send: &mpsc::Sender<ClientRequest>,
-    ui: &mut UIStateGuard,
     state: &SharedState,
 ) -> Result<bool> {
-    match state.player.read().unwrap().context() {
+    let uri = match state.ui.lock().current_page().context_uri() {
+        Some(uri) => uri,
+        None => return Ok(false),
+    };
+
+    match state.data.read().caches.context.peek(&uri) {
         Some(context) => match context {
-            Context::Artist(_, ref tracks, ref albums, ref artists) => {
-                let focus_state = match ui.window {
-                    WindowState::Artist(_, _, _, state) => state,
-                    _ => unreachable!(),
+            Context::Artist {
+                top_tracks,
+                albums,
+                related_artists,
+                ..
+            } => {
+                let focus_state = match state.ui.lock().window {
+                    WindowState::Artist { focus, .. } => focus,
+                    _ => return Ok(false),
                 };
 
                 match focus_state {
                     ArtistFocusState::Albums => handle_command_for_album_list_subwindow(
                         command,
-                        send,
-                        ui,
-                        ui.filtered_items_by_search(albums),
+                        state,
+                        state.filtered_items_by_search(albums),
                     ),
                     ArtistFocusState::RelatedArtists => handle_command_for_artist_list_subwindow(
                         command,
-                        send,
-                        ui,
-                        ui.filtered_items_by_search(artists),
+                        state,
+                        state.filtered_items_by_search(related_artists),
                     ),
                     ArtistFocusState::TopTracks => handle_command_for_track_table_subwindow(
                         command,
                         send,
-                        ui,
+                        state,
                         None,
-                        Some(tracks.iter().map(|t| &t.id).collect()),
-                        ui.filtered_items_by_search(tracks),
+                        Some(top_tracks.iter().map(|t| &t.id).collect()),
+                        state.filtered_items_by_search(top_tracks),
                     ),
                 }
             }
-            Context::Album(ref album, ref tracks) => handle_command_for_track_table_subwindow(
+            Context::Album { album, tracks } => handle_command_for_track_table_subwindow(
                 command,
                 send,
-                ui,
+                state,
                 Some(ContextId::Album(album.id.clone())),
                 None,
-                ui.filtered_items_by_search(tracks),
+                state.filtered_items_by_search(tracks),
+            ),
+            Context::Playlist { playlist, tracks } => handle_command_for_track_table_subwindow(
+                command,
+                send,
+                state,
+                Some(ContextId::Playlist(playlist.id.clone())),
+                None,
+                state.filtered_items_by_search(tracks),
             ),
-            Context::Playlist(ref playlist, ref tracks) => {
-                handle_command_for_track_table_subwindow(
-                    command,
-                    send,
-                    ui,
-                    Some(ContextId::Playlist(playlist.id.clone())),
-                    None,
-                    ui.filtered_items_by_search(tracks),
-                )
-            }
         },
         None => Ok(false),
     }
@@ -310,12 +406,13 @@ pub fn handle_command_for_focused_context_subwindow(
 fn handle_command_for_track_table_subwindow(
     command: Command,
     send: &mpsc::Sender<ClientRequest>,
-    ui: &mut UIStateGuard,
+    state: &SharedState,
     context_id: Option<ContextId>,
     track_ids: Option<Vec<&TrackId>>,
     tracks: Vec<&Track>,
 ) -> Result<bool> {
-    let id = ui.window.selected().unwrap();
+    let mut ui = state.ui.lock();
+    let id = ui.window.selected().unwrap_or_default();
 
     match command {
         Command::SelectNextOrScrollDown => {
@@ -329,7 +426,7 @@ fn handle_command_for_track_table_subwindow(
             }
         }
         Command::ChooseSelected => {
-            let offset = Some(model::Offset::for_uri(&tracks[id].id.uri()));
+            let offset = Some(rspotify_model::Offset::for_uri(&tracks[id].id.uri()));
             if track_ids.is_some() {
                 // play a track from a list of tracks
                 send.send(ClientRequest::Player(PlayerRequest::StartPlayback(
@@ -356,10 +453,11 @@ fn handle_command_for_track_table_subwindow(
 fn handle_command_for_track_list_subwindow(
     command: Command,
     send: &mpsc::Sender<ClientRequest>,
-    ui: &mut UIStateGuard,
+    state: &SharedState,
     tracks: Vec<&Track>,
 ) -> Result<bool> {
-    let id = ui.window.selected().unwrap();
+    let mut ui = state.ui.lock();
+    let id = ui.window.selected().unwrap_or_default();
 
     match command {
         Command::SelectNextOrScrollDown => {
@@ -395,11 +493,11 @@ fn handle_command_for_track_list_subwindow(
 
 fn handle_command_for_artist_list_subwindow(
     command: Command,
-    send: &mpsc::Sender<ClientRequest>,
-    ui: &mut UIStateGuard,
+    state: &SharedState,
     artists: Vec<&Artist>,
 ) -> Result<bool> {
-    let id = ui.window.selected().unwrap();
+    let mut ui = state.ui.lock();
+    let id = ui.window.selected().unwrap_or_default();
 
     match command {
         Command::SelectNextOrScrollDown => {
@@ -414,8 +512,10 @@ fn handle_command_for_artist_list_subwindow(
         }
         Command::ChooseSelected => {
             let context_id = ContextId::Artist(artists[id].id.clone());
-            send.send(ClientRequest::GetContext(context_id.clone()))?;
-            ui.new_page(PageState::Browsing(context_id));
+            ui.create_new_page(PageState::Context(
+                None,
+                ContextPageType::Browsing(context_id),
+            ));
         }
         Command::ShowActionsOnSelectedItem => {
             ui.popup = Some(PopupState::ActionList(
@@ -430,11 +530,11 @@ fn handle_command_for_artist_list_subwindow(
 
 fn handle_command_for_album_list_subwindow(
     command: Command,
-    send: &mpsc::Sender<ClientRequest>,
-    ui: &mut UIStateGuard,
+    state: &SharedState,
     albums: Vec<&Album>,
 ) -> Result<bool> {
-    let id = ui.window.selected().unwrap();
+    let mut ui = state.ui.lock();
+    let id = ui.window.selected().unwrap_or_default();
 
     match command {
         Command::SelectNextOrScrollDown => {
@@ -449,8 +549,10 @@ fn handle_command_for_album_list_subwindow(
         }
         Command::ChooseSelected => {
             let context_id = ContextId::Album(albums[id].id.clone());
-            send.send(ClientRequest::GetContext(context_id.clone()))?;
-            ui.new_page(PageState::Browsing(context_id));
+            ui.create_new_page(PageState::Context(
+                None,
+                ContextPageType::Browsing(context_id),
+            ));
         }
         Command::ShowActionsOnSelectedItem => {
             ui.popup = Some(PopupState::ActionList(
@@ -465,11 +567,11 @@ fn handle_command_for_album_list_subwindow(
 
 fn handle_command_for_playlist_list_subwindow(
     command: Command,
-    send: &mpsc::Sender<ClientRequest>,
-    ui: &mut UIStateGuard,
+    state: &SharedState,
     playlists: Vec<&Playlist>,
 ) -> Result<bool> {
-    let id = ui.window.selected().unwrap();
+    let mut ui = state.ui.lock();
+    let id = ui.window.selected().unwrap_or_default();
 
     match command {
         Command::SelectNextOrScrollDown => {
@@ -484,8 +586,10 @@ fn handle_command_for_playlist_list_subwindow(
         }
         Command::ChooseSelected => {
             let context_id = ContextId::Playlist(playlists[id].id.clone());
-            send.send(ClientRequest::GetContext(context_id.clone()))?;
-            ui.new_page(PageState::Browsing(context_id));
+            ui.create_new_page(PageState::Context(
+                None,
+                ContextPageType::Browsing(context_id),
+            ));
         }
         Command::ShowActionsOnSelectedItem => {
             ui.popup = Some(PopupState::ActionList(
