@@ -1,5 +1,5 @@
 use super::Frame;
-use crate::{config, state::*};
+use crate::state::*;
 use std::collections::{btree_map::Entry, BTreeMap};
 use tui::{layout::*, widgets::*};
 
@@ -15,30 +15,29 @@ const COMMAND_TABLE_CONSTRAINTS: [Constraint; 3] = [
     Constraint::Percentage(50),
 ];
 
-/// renders a shortcut help window to show the available shortcuts
-/// based on user's inputs
-pub fn render_shortcut_help_window(
-    frame: &mut Frame,
-    ui: &UIStateGuard,
-    state: &SharedState,
-    rect: Rect,
-) -> Rect {
+/// renders a shortcut help popup to show the available shortcuts based on user's inputs
+pub fn render_shortcut_help_popup(frame: &mut Frame, state: &SharedState, rect: Rect) -> Rect {
+    let ui = state.ui.lock();
     let input = &ui.input_key_sequence;
-    // render the shortcuts help table if needed
-    let matches = if input.keys.is_empty() {
-        vec![]
-    } else {
-        state
-            .keymap_config
-            .find_matched_prefix_keymaps(input)
-            .into_iter()
-            .map(|keymap| {
-                let mut keymap = keymap.clone();
-                keymap.key_sequence.keys.drain(0..input.keys.len());
-                keymap
-            })
-            .filter(|keymap| !keymap.key_sequence.keys.is_empty())
-            .collect::<Vec<_>>()
+
+    // get the matches (keymaps) from the current key sequence input,
+    // if there is at lease one match, render the shortcut help popup
+    let matches = {
+        if input.keys.is_empty() {
+            vec![]
+        } else {
+            state
+                .keymap_config
+                .find_matched_prefix_keymaps(input)
+                .into_iter()
+                .map(|keymap| {
+                    let mut keymap = keymap.clone();
+                    keymap.key_sequence.keys.drain(0..input.keys.len());
+                    keymap
+                })
+                .filter(|keymap| !keymap.key_sequence.keys.is_empty())
+                .collect::<Vec<_>>()
+        }
     };
 
     if matches.is_empty() {
@@ -48,21 +47,34 @@ pub fn render_shortcut_help_window(
             .direction(Direction::Vertical)
             .constraints([Constraint::Min(0), Constraint::Length(7)].as_ref())
             .split(rect);
-        render_shortcuts_help_window(matches, frame, ui, chunks[1]);
+
+        let help_table = Table::new(
+            matches
+                .into_iter()
+                .map(|km| format!("{}: {:?}", km.key_sequence, km.command))
+                .collect::<Vec<_>>()
+                .chunks(SHORTCUT_TABLE_N_COLUMNS)
+                .map(|c| Row::new(c.iter().map(|i| Cell::from(i.to_owned()))))
+                .collect::<Vec<_>>(),
+        )
+        .widths(&SHORTCUT_TABLE_CONSTRAINS)
+        .block(
+            Block::default()
+                .title(ui.theme.block_title_with_style("Shortcuts"))
+                .borders(Borders::ALL),
+        );
+        frame.render_widget(help_table, chunks[1]);
         chunks[0]
     }
 }
 
-/// renders a command help window listing all key shortcuts and corresponding descriptions
-pub fn render_commands_help_window(
-    frame: &mut Frame,
-    ui: &mut UIStateGuard,
-    state: &SharedState,
-    rect: Rect,
-) {
+/// renders a command help popup listing all key shortcuts and corresponding descriptions
+pub fn render_commands_help_popup(frame: &mut Frame, state: &SharedState, rect: Rect) {
+    let mut ui = state.ui.lock();
+
     let offset = match ui.popup {
-        Some(PopupState::CommandHelp(ref mut offset)) => offset,
-        _ => unreachable!(),
+        Some(PopupState::CommandHelp { ref mut offset }) => offset,
+        _ => return,
     };
 
     let mut map = BTreeMap::new();
@@ -101,7 +113,7 @@ pub fn render_commands_help_window(
             Cell::from("Shortcuts"),
             Cell::from("Description"),
         ])
-        .style(ui.theme.context_tracks_table_header()),
+        .style(ui.theme.table_header()),
     )
     .widths(&COMMAND_TABLE_CONSTRAINTS)
     .block(
@@ -111,28 +123,3 @@ pub fn render_commands_help_window(
     );
     frame.render_widget(help_table, rect);
 }
-
-/// renders a shortcuts help widget from a list of keymaps
-fn render_shortcuts_help_window(
-    keymaps: Vec<config::Keymap>,
-    frame: &mut Frame,
-    ui: &UIStateGuard,
-    rect: Rect,
-) {
-    let help_table = Table::new(
-        keymaps
-            .into_iter()
-            .map(|km| format!("{}: {:?}", km.key_sequence, km.command))
-            .collect::<Vec<_>>()
-            .chunks(SHORTCUT_TABLE_N_COLUMNS)
-            .map(|c| Row::new(c.iter().map(|i| Cell::from(i.to_owned()))))
-            .collect::<Vec<_>>(),
-    )
-    .widths(&SHORTCUT_TABLE_CONSTRAINS)
-    .block(
-        Block::default()
-            .title(ui.theme.block_title_with_style("Shortcuts"))
-            .borders(Borders::ALL),
-    );
-    frame.render_widget(help_table, rect);
-}
